/*	rdwr_popen.c
 *	S.Coffin   GV Computing  3/91
 *
 *	rdwr_popen: like popen, but return 3 FILEs, for in, out & err
 *	NOTE: you usually want to change the buffering on the returned
 *	FILEs (after they've been opened), probably via a setbuf( fp, NULL )
 *
 *	original for email-II	S.Coffin  7/86
 */

/*  Copyright (c) 1992,1993,1994,1995 by Stephen Coffin.  All rights reserved.
 *
 *  This program is distributed in the hope that it will be useful.
 *  Use and copying of this software and preparation of derivative works
 *  based upon this software are permitted, so long as the following
 *  conditions are met:
 *       o credit to the authors is acknowledged following current
 *         academic behavior
 *       o no fees or compensation are charged for use, copies, or
 *         access to this software
 *	 o no fees or compensation are charged for access to the
 *	   game server for which this software is a client
 *       o this copyright notice is included intact.
 *  This software is made available AS IS, and no warranty is made about 
 *  the software or its performance. 
 * 
 *  Bug descriptions, use reports, comments or suggestions are welcome.
 *  Send them to    scoffin@netcom.com
 */

#include <stdio.h>
#include "filter.h"

extern FILE *fdopen();
extern void exit ();
extern unsigned sleep ();
static void my_exec();

static int pid;

/* returns the pid we exec'ed or zero on error */
int rdwr_popen (fr, fw, ferr, cmd, env, envdelim )
	FILE **fr, **fw, **ferr;
	char *cmd, *env;
	char envdelim;
{
	int tries = 5;
	unsigned ZZs = 1;
	int fda [2], fdb [2], fdc [2];
	int i;
	char *e, *t;

	if( dbg ) fprintf( stderr, "in rdwr_popen  cmd=>>%s<<\n", cmd );

	/* Get all the pipes we need */
	if( pipe (fda) || pipe (fdb) || pipe (fdc) ) return( 0 );
		
	/* Try & fork 5 times, backing off 1, 2, 4 .. seconds each try */
	while ((pid = fork ()) < 0) {
		if (--tries == 0) return( 0 );
		(void) sleep (ZZs);
		ZZs <<= 1;
	}
	if (pid == 0) {			/* child */
		/* add the envstr to the real environment */
/*SC  xigc does not use the env arg; we comment this out since some
 *SC  systems do not have "putenv()"
		for( e = env ; e != NULL && e[0] != '\0'; e = ++t ) {
		    t = strchr( e, envdelim );
		    if( t != NULL ) *t = '\0';
		    if( dbg ) {
			fprintf( stderr, "env=>>%s<<\n", e );
			fflush( stderr );
		    }
		    putenv( e );
		    if( dbg ) {
			fprintf( stderr, "PUTENV DONE\n" );
			fflush( stderr );
		    }
		    if( t == NULL ) break;
		}
SC*/

		if( fda[0] != 0 ) {
			(void) close( 0 );
			i = dup (fda [0]);
			if( i != 0 ) exit( (-1) );
		}
		if( fdb[1] != 1 ) {
			(void)close (1);
			if (dup (fdb [1]) != 1) exit( (-1) );
		}
		if( fdc[1] != 2 ) {
			(void)close (2);
			if (dup (fdc [1]) != 2) exit( (-1) );
		}
			
		/* Make sure there are no inherited file descriptors */
		for (i = 3; i < NOFILE; i += 1) (void) close (i);

		/* and exec */
		my_exec( cmd );
		exit( 0 );
		/*NOTREACHED*/
	} else {
		/* Parent */
		if (fr && (*fr = fdopen (fdb [0], "r")) == NULL) return( (-1) );
		if (fw && (*fw = fdopen (fda [1], "w")) == NULL) return( (-1) );
		if (ferr && (*ferr = fdopen (fdc [0], "r")) == NULL) return( (-1) );
		(void) close (fdc [1]);
		(void) close (fdb [1]);
		(void) close (fda [0]);
		return( pid );
	}
}

/*ARGSUSED*/
rdwr_pclose( fin, fout, ferr )
	FILE *fin, *fout, *ferr;
{
	int ret;

	ret = kill( pid, SIGKILL );
	if( ret == 0 ) wait( (int *)0 );
}

static void my_exec (ccmd)
	char *ccmd;
{

/* Can only be MAXARGS arguments */
#define MAXARGS	64

	char *argv [MAXARGS];
	register char **argv_p, *cp;

	if( dbg ) {
	    fprintf( stderr, "my_exec  cmd=>>%s<<\n", ccmd );
	    fflush( stderr );
	}
	
	argv_p = argv;
	cp = ccmd;
	/* Scan up cmd, splitting arguments into argv. This is the
	 * child, so we can zap things in cmd safely */
	
	while (*cp != '\0') {
		/* Skip any white space */
		while (*cp && isspace (*cp)) cp += 1;
		if (*cp == '\0') break;
		*argv_p++ = cp;
		while (*cp && !isspace (*cp)) cp += 1;
		if (*cp == '\0') break;
		*cp++ = '\0';
	}
	*argv_p = NULL;
	execvp (ccmd, argv);
}
